@@ -1,7 +1,6 @@
 //! Stateful Base64 decoder.
 
 use crate::{
-    encoding::decode_padding,
     variant::Variant,
     Encoding,
     Error::{self, InvalidLength},
@@ -46,7 +45,7 @@ impl<'i, E: Variant> Decoder<'i, E> {
     /// (non-newline-delimited) Base64-encoded data.
     pub fn new(input: &'i [u8]) -> Result<Self, Error> {
         Ok(Self {
-            line: Line::new(Self::unpad_input(input)?),
+            line: Line::new(input),
             line_reader: LineReader::default(),
             block_buffer: BlockBuffer::default(),
             encoding: PhantomData,
@@ -75,7 +74,7 @@ impl<'i, E: Variant> Decoder<'i, E> {
     pub fn new_wrapped(input: &'i [u8], line_width: usize) -> Result<Self, Error> {
         Ok(Self {
             line: Line::default(),
-            line_reader: LineReader::new(Self::unpad_input(input)?, line_width)?,
+            line_reader: LineReader::new(input, line_width)?,
             block_buffer: BlockBuffer::default(),
             encoding: PhantomData,
         })
@@ -93,15 +92,15 @@ impl<'i, E: Variant> Decoder<'i, E> {
             return Err(Error::InvalidLength);
         }
 
-        let mut out_off = 0;
+        let mut out_pos = 0;
 
-        while out_off < out.len() {
+        while out_pos < out.len() {
             // If there's data in the block buffer, use it
             if !self.block_buffer.is_empty() {
-                let out_rem = out.len().checked_sub(out_off).ok_or(InvalidLength)?;
+                let out_rem = out.len().checked_sub(out_pos).ok_or(InvalidLength)?;
                 let bytes = self.block_buffer.take(out_rem);
-                out[out_off..][..bytes.len()].copy_from_slice(bytes);
-                out_off = out_off.checked_add(bytes.len()).ok_or(InvalidLength)?;
+                out[out_pos..][..bytes.len()].copy_from_slice(bytes);
+                out_pos = out_pos.checked_add(bytes.len()).ok_or(InvalidLength)?;
             }
 
             // Advance the line reader if necessary
@@ -111,18 +110,18 @@ impl<'i, E: Variant> Decoder<'i, E> {
 
             // Attempt to decode a stride of block-aligned data
             let in_blocks = self.line.len() / 4;
-            let out_rem = out.len().checked_sub(out_off).ok_or(InvalidLength)?;
+            let out_rem = out.len().checked_sub(out_pos).ok_or(InvalidLength)?;
             let out_blocks = out_rem / 3;
             let blocks = cmp::min(in_blocks, out_blocks);
             let in_aligned = self.line.take(blocks * 4);
 
             if !in_aligned.is_empty() {
-                let out_buf = &mut out[out_off..][..(blocks * 3)];
-                let decoded_len = E::Unpadded::decode(in_aligned, out_buf)?.len();
-                out_off = out_off.checked_add(decoded_len).ok_or(InvalidLength)?;
+                let out_buf = &mut out[out_pos..][..(blocks * 3)];
+                let decoded_len = self.perform_decode(in_aligned, out_buf)?.len();
+                out_pos = out_pos.checked_add(decoded_len).ok_or(InvalidLength)?;
             }
 
-            if out_off < out.len() {
+            if out_pos < out.len() {
                 // If we still haven't filled the output slice, we're in a
                 // situation where either the input or output isn't
                 // block-aligned, so fill the internal block buffer
@@ -140,7 +139,9 @@ impl<'i, E: Variant> Decoder<'i, E> {
 
     /// Fill the block buffer with data.
     fn fill_block_buffer(&mut self) -> Result<(), Error> {
-        if self.line.len() < 4 && !self.line_reader.is_empty() {
+        let mut buf = [0u8; BlockBuffer::SIZE];
+
+        let decoded = if self.line.len() < 4 && !self.line_reader.is_empty() {
             // Handle input block which is split across lines
             let mut tmp = [0u8; 4];
 
@@ -158,11 +159,13 @@ impl<'i, E: Variant> Decoder<'i, E> {
                 .checked_add(line_end.len())
                 .ok_or(InvalidLength)?;
 
-            self.block_buffer.fill::<E::Unpadded>(&tmp[..tmp_len])
+            self.perform_decode(&tmp[..tmp_len], &mut buf)
         } else {
             let block = self.line.take(4);
-            self.block_buffer.fill::<E::Unpadded>(block)
-        }
+            self.perform_decode(block, &mut buf)
+        }?;
+
+        self.block_buffer.fill(decoded)
     }
 
     /// Advance the internal buffer to the next line.
@@ -177,22 +180,12 @@ impl<'i, E: Variant> Decoder<'i, E> {
         }
     }
 
-    /// Remove padding from an input buffer.
-    // TODO(tarcieri): instead of this, process the last Base64 block as padded?
-    // This approach may not cover all cases with linewrapped Base64
-    fn unpad_input(input: &[u8]) -> Result<&[u8], Error> {
-        if E::PADDED {
-            // TODO(tarcieri): validate that padding is well-formed with `validate_padding`
-            // ...or switch to processing the last block as padded, leaning on
-            // the existing padding validation code
-            let (unpadded_len, err) = decode_padding(input)?;
-            if err != 0 {
-                return Err(Error::InvalidEncoding);
-            }
-
-            Ok(&input[..unpadded_len])
+    /// Perform Base64 decoding operation.
+    fn perform_decode<'o>(&self, src: &[u8], dst: &'o mut [u8]) -> Result<&'o [u8], Error> {
+        if self.is_finished() {
+            E::decode(src, dst)
         } else {
-            Ok(input)
+            E::Unpadded::decode(src, dst)
         }
     }
 }
@@ -201,10 +194,10 @@ impl<'i, E: Variant> Decoder<'i, E> {
 ///
 /// This handles a partially decoded block of data, i.e. data which has been
 /// decoded but not read.
-#[derive(Clone, Default)]
+#[derive(Clone, Default, Debug)]
 struct BlockBuffer {
     /// 3 decoded bytes from a 4-byte Base64-encoded input.
-    decoded: [u8; 3],
+    decoded: [u8; Self::SIZE],
 
     /// Length of the buffer.
     length: usize,
@@ -214,12 +207,20 @@ struct BlockBuffer {
 }
 
 impl BlockBuffer {
-    /// Fill the buffer by decoding up to 4 bytes of Base64 input
-    fn fill<E: Variant>(&mut self, base64_input: &[u8]) -> Result<(), Error> {
+    /// Size of the buffer in bytes.
+    const SIZE: usize = 3;
+
+    /// Fill the buffer by decoding up to 3 bytes of decoded Base64 input.
+    fn fill(&mut self, decoded_input: &[u8]) -> Result<(), Error> {
         debug_assert!(self.is_empty());
-        debug_assert!(base64_input.len() <= 4);
-        self.length = E::decode(base64_input, &mut self.decoded)?.len();
+
+        if decoded_input.len() > Self::SIZE {
+            return Err(Error::InvalidLength);
+        }
+
         self.position = 0;
+        self.length = decoded_input.len();
+        self.decoded[..decoded_input.len()].copy_from_slice(decoded_input);
         Ok(())
     }
 
@@ -247,7 +248,7 @@ impl BlockBuffer {
 }
 
 /// A single line of linewrapped data, providing a read buffer.
-#[derive(Clone)]
+#[derive(Clone, Debug)]
 pub struct Line<'i> {
     /// Remaining data in the line
     remaining: &'i [u8],
@@ -323,7 +324,7 @@ impl<'i> Iterator for LineReader<'i> {
     fn next(&mut self) -> Option<Result<Line<'i>, Error>> {
         if let Some(line_width) = self.line_width {
             let rest = match self.remaining.get(line_width..) {
-                None => {
+                None | Some([]) => {
                     if self.remaining.is_empty() {
                         return None;
                     } else {
@@ -335,7 +336,10 @@ impl<'i> Iterator for LineReader<'i> {
                 Some([CHAR_CR, CHAR_LF, rest @ ..]) => rest,
                 Some([CHAR_CR, rest @ ..]) => rest,
                 Some([CHAR_LF, rest @ ..]) => rest,
-                _ => return Some(Err(Error::InvalidEncoding)),
+                _ => {
+                    // Expected a leading newline
+                    return Some(Err(Error::InvalidEncoding));
+                }
             };
 
             let line = Line::new(&self.remaining[..line_width]);
